#pragma once

#include "list.hpp"
#include "dict.hpp"
#include "if.hpp"

// drop loop
template<typename Program, int Depth>
struct drop_loop_impl;

template<typename Program>
struct drop_loop_impl<
	cons<']', Program>,
	0
	> {
	typedef Program value;
};

template<typename Program, int Depth>
struct drop_loop_impl<
	cons<'[', Program>,
	Depth
	> {
	typedef typename drop_loop_impl<Program, Depth + 1>::value value;
};

template<typename Program, int Depth>
struct drop_loop_impl<
	cons<']', Program>,
	Depth
	> {
	typedef typename drop_loop_impl<Program, Depth - 1>::value value;
};

template<typename Program, int Depth, int c>
struct drop_loop_impl<
	cons<c, Program>,
	Depth
	> {
	typedef typename drop_loop_impl<Program, Depth>::value value;
};


template<typename Program>
struct drop_loop {
	typedef typename drop_loop_impl<Program, 0>::value value;
};


// take loop
template<typename Program, int Depth>
struct take_loop_impl;

template<typename Program>
struct take_loop_impl<
	cons<']', Program>,
	0
	> {
	typedef nil value;
};

template<typename Program, int Depth, int c>
struct take_loop_impl<
	cons<c, Program>,
	Depth
	> {
	typedef cons<c, typename take_loop_impl<Program, Depth>::value> value;
};

template<typename Program, int Depth>
struct take_loop_impl<
	cons<'[', Program>,
	Depth
	> {
	typedef cons<'[', typename take_loop_impl<Program, Depth + 1>::value> value;
};

template<typename Program, int Depth>
struct take_loop_impl<
	cons<']', Program>,
	Depth
	> {
	typedef cons<']', typename take_loop_impl<Program, Depth - 1>::value> value;
};

template<typename Program>
struct take_loop {
	typedef typename take_loop_impl<Program, 0>::value value;
};


// interpret

template<typename Program, typename Cells, int Pointer>
struct interpret;

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'.', Program>,
	Cells,
	Pointer
	> {
	typedef cons<
		at_key<Cells, Pointer>::value,
		typename interpret<Program, Cells, Pointer>::value
	> value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'>', Program>,
	Cells,
	Pointer
	> {
	typedef typename interpret<Program, Cells, Pointer + 1>::value value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'<', Program>,
	Cells,
	Pointer
	> {
	typedef typename interpret<Program, Cells, Pointer - 1>::value value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'+', Program>,
	Cells,
	Pointer
	> {
	typedef typename interpret<
		Program,
		typename set_value<
			Cells,
			Pointer,
			at_key<Cells, Pointer>::value + 1
		>::value,
		Pointer
	>::value value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'-', Program>,
	Cells,
	Pointer
	> {
	typedef typename interpret<
		Program,
		typename set_value<
			Cells,
			Pointer,
			at_key<Cells, Pointer>::value - 1
		>::value,
		Pointer
	>::value value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<'[', Program>,
	Cells,
	Pointer
	> {
	typedef typename IF<
		(at_key<Cells, Pointer>::value > 0),
		interpret<
			typename concat<
				typename take_loop<Program>::value,
				cons<'[', Program>
			>::value,
			Cells,
			Pointer
		>,
		interpret<
			typename drop_loop<Program>::value,
			Cells,
			Pointer
		>
	>::value::value value;
};

template<typename Program, typename Cells, int Pointer>
struct interpret<
	cons<']', Program>,
	Cells,
	Pointer
	> {
	typedef typename interpret<Program, Cells, Pointer>::value value;
};

template<typename Cells, int Pointer>
struct interpret<nil, Cells, Pointer> {
	typedef nil value;
};
