#include <iostream>
#include <typeinfo>
#include <string>
#include <stdlib.h>
#include <cxxabi.h>
#include <stdint.h>

class Expression {
	public:
		virtual void print(int values[]) = 0;
};

class Element : public Expression {
	private:
		int element;

	public:
		Element(int element) : element (element) {}

		void print(int values[]) {
			std::cout << values[element];
		}
};

template<class ExpL, class ExpR, char Oper>
class Operation : public Expression {
	private:
		ExpL l;
		ExpR r;

	public:
		Operation(ExpL l, ExpR r) : l(l), r(r) {}

		void print(int values[]) {
			l.print(values);
			std::cout << Oper;
			r.print(values);
		}
};

template<class ExpL, class ExpR>
Operation<ExpL, ExpR, '*'> operator*(ExpL l, ExpR r)
{
	return Operation<ExpL, ExpR, '*'>(l, r);
}

template<class ExpL, class ExpR>
Operation<ExpL, ExpR, '+'> operator+(ExpL l, ExpR r)
{
	return Operation<ExpL, ExpR, '+'>(l, r);
}

class ExpressionHandle : public Expression {
	private:
		Expression *handled;

	public:
		template<class Exp>
		ExpressionHandle(Exp exp) : handled(new Exp(exp)) {}

		void print(int values[])
		{
			handled->print(values);
			char *realname = abi::__cxa_demangle(typeid(*handled).name(), 0, 0, 0);
			std::cout << std::endl << realname << std::endl;
			free(realname);
		}
};

int main()
{
	ExpressionHandle exp = Element(1) * Element(2) + Element(3) * Element(1);
	int values[4] = { 0, 3, 1, 2 };
	exp.print(values);
}
